/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 * 
 * Copyright (©) 2005,2006,2007 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * Copyright (©) 2002,2007 Mark Wedel & Crossfire Development Team
 * Copyright (©) 1992,2007 Frank Tore Johansen
 * 
 * Deliantra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

#include <global.h>
#include <random_map.h>
#include <rproto.h>

/* some monsters are multisquare, and these guys require special
        handling. */

void
insert_multisquare_ob_in_map (object *new_obj, maptile *map)
{
  int x, y;
  object *old_seg;
  object *head;

  /* first insert the head */
  insert_ob_in_map (new_obj, map, new_obj, INS_NO_MERGE | INS_NO_WALK_ON);

  x = new_obj->x;
  y = new_obj->y;
  old_seg = new_obj;
  head = new_obj;

  //TODO: use expand_tail + insert instead
  for (archetype *at = (archetype *)new_obj->arch->more; at; at = (archetype *)at->more)
    {
      object *new_seg;

      new_seg = arch_to_object (at);
      new_seg->x = x + at->x;
      new_seg->y = y + at->y;
      new_seg->map = old_seg->map;
      insert_ob_in_map (new_seg, new_seg->map, new_seg, INS_NO_MERGE | INS_NO_WALK_ON);
      new_seg->head = head;
      old_seg->more = new_seg;
      old_seg = new_seg;
    }

  old_seg->more = 0;
}

/*  place some monsters into the map. */
void
place_monsters (maptile *map, char *monsterstyle, int difficulty, random_map_params *RP)
{
  char styledirname[1024];
  maptile *style_map = 0;
  int failed_placements;
  sint64 exp_per_sq, total_experience;
  int number_monsters = 0;
  archetype *at;

  sprintf (styledirname, "%s", "/styles/monsterstyles");
  style_map = find_style (styledirname, monsterstyle, difficulty);
  if (!style_map)
    return;

  /* fill up the map with random monsters from the monster style */

  total_experience = 0;
  failed_placements = 0;
  exp_per_sq = 0;
  while (exp_per_sq <= level_exp (difficulty, 1.0) && failed_placements < 100 && number_monsters < (RP->Xsize * RP->Ysize) / 8)
    {
      object *this_monster = style_map->pick_random_object ();
      int x, y, freeindex;

      if (this_monster == NULL)
        return;                 /* no monster?? */

      x = rndm (RP->Xsize);
      y = rndm (RP->Ysize);
      freeindex = find_first_free_spot (this_monster, map, x, y);
      if (freeindex != -1)
        {
          object *new_monster = arch_to_object (this_monster->arch);

          x += freearr_x[freeindex];
          y += freearr_y[freeindex];
          copy_object_with_inv (this_monster, new_monster);
          new_monster->x = x;
          new_monster->y = y;
          insert_multisquare_ob_in_map (new_monster, map);

          if (new_monster->is_alive ())
            {
              total_experience += this_monster->stats.exp;

              for (at = new_monster->arch; at; at = (archetype *)at->more)
                number_monsters++;

              assert (new_monster->stats.hp >= 0);
              RP->total_map_hp += new_monster->stats.hp;    /*  a global count */
            }
          else
            failed_placements++;
        }
      else
        failed_placements++;

      exp_per_sq = (sint64) (((double) 1000 * total_experience) / (map->width * map->height + 1));
    }
}
